""" This script is used for writing in HTML files

It adds links to HTML table.
It generates span tags for un/colored matching blocks.
It compares two text files
It inserts comparison results in corresponding html files

"""

from os import fsync, rename, path
from random import randint
from shutil import copy
from typing import Any, List

from bs4 import BeautifulSoup as Bs
from tabulate import tabulate

from scripts.html_utils import (
    get_color_from_similarity,
    get_real_matching_blocks,
    blocks_list_to_strings_list,
    get_ordered_blocks_positions,
)
from scripts.utils import is_float


def add_links_to_html_table(html_path: str) -> None:
    """Add links to HTML data cells at specified path

    This method will link to all HTML TD tags which contain a float different from - 1 the
    corresponding HTML comparison file. The links will be opened in a new tab. The colors of
    the text in tag will change depending on similarity score.

    """

    with open(html_path, encoding="utf-8") as html:
        soup = Bs(html, "html.parser")
        file_ind = 0  # Cursor on file number for the naming of html files

        for td_tag in soup.findAll("td"):  # Retrieve all data celss from html table in path
            if is_float(td_tag.text):  # If td is not filename or -1
                tmp = soup.new_tag(
                    "a",
                    href="file:///" + html_path.replace("_results", str(file_ind)),
                    target="_blank",
                    style="color:" + get_color_from_similarity(td_tag.text),
                )

                td_tag.string.wrap(tmp)  # We wrap the td string between the hyperlink
                file_ind += 1

        # We update the HTML of the file at path
        with open(html_path, "wb") as f_output:
            f_output.write(soup.prettify("utf-8"))
            f_output.flush()
            fsync(f_output.fileno())
            f_output.close()


def get_span_blocks(bs_obj: Bs, text1: list, text2: list, block_size: int) -> list:
    """Return list of spans with colors for HTML rendering"""

    results: List[List[Any]] = [[], []]  # List of spans list

    # Get matching blocks with chosen minimum size
    matching_blocks = get_real_matching_blocks(text1, text2, block_size)

    # Generate one unique color for each matching block
    colors = [f"#{randint(0, 0xFFFFFF):06X}" for _ in range(len(matching_blocks))]

    # Convert blocks from list of list of strings to list of strings
    string_blocks = [" ".join(map(str, text1[b.a : b.a + b.size])) for b in matching_blocks]

    # Store lengths of blocks in text
    strings_len_list = blocks_list_to_strings_list(matching_blocks, text1)

    # Convert list of strings to strings
    str1, str2 = " ".join(map(str, text1)), " ".join(map(str, text2))

    global_positions_list = [
        get_ordered_blocks_positions(str1, matching_blocks, string_blocks),
        get_ordered_blocks_positions(str2, matching_blocks, string_blocks),
    ]

    for num, pos_list in enumerate(global_positions_list):
        cursor = 0  # Cursor on current string

        if num == 1:  # Second iteration on second string
            str1 = str2

        for block in pos_list:
            # Span tag for the text before the matching sequence
            span = bs_obj.new_tag("span")
            span.string = str1[cursor : block[0]]

            # Span tag for the text in the matching sequence
            blockspan = bs_obj.new_tag("span", style="color:" + colors[block[1]] + "; font-weight:bold")
            blockspan.string = str1[block[0] : block[0] + strings_len_list[block[1]]]

            # Append spans tags to results list
            results[num].append(span)
            results[num].append(blockspan)

            # Update cursor position after last matching sequence
            cursor = block[0] + strings_len_list[block[1]]

        # End of loop, last span tag for the rest of the text
        span = bs_obj.new_tag("span")
        span.string = str1[cursor:]
        results[num].append(span)

    return results


def papers_comparison(save_dir: str, ind: int, text1: list, text2: list, filenames: tuple, block_size: int) -> None:
    """Write to HTML file texts that have been compared with highlighted similar blocks"""

    copy(path.join("templates", "template.html"), save_dir)  # Copy comparison template to curr dir
    comp_path = path.join(save_dir, str(ind) + ".html")
    rename(path.join(save_dir, "template.html"), comp_path)

    with open(comp_path, encoding="utf-8") as html:
        soup = Bs(html, "html.parser")
        res = get_span_blocks(soup, text1, text2, block_size)
        blocks = soup.findAll(attrs={"class": "block"})

        # Append filename tags and span tags to html
        for i, filename in enumerate(filenames):
            temp_tag = soup.new_tag("h3")
            temp_tag.string = filename
            blocks[i].append(temp_tag)
            for tag in res[i]:
                blocks[i].append(tag)

    with open(comp_path, "wb") as f_output:
        f_output.write(soup.prettify("utf-8"))


def results_to_html(scores: list, files_names: list, html_path: str) -> None:
    """Write similarity results to HTML page"""

    for ind, file_name in enumerate(files_names):
        scores[ind].insert(0, file_name)

    scores.insert(0, files_names)
    scores[0].insert(0, "")

    with open(html_path, "w", encoding="utf-8") as file:
        file.write(tabulate(scores, tablefmt="html"))
        file.flush()
        fsync(file.fileno())
        file.close()



""" This function calculates similarity scores with different methods

It calculates similarity scores with :
- difflib library to find matching sequences.
- Jaccard Similarity
- words counting,
- overlapping words

"""

import difflib

from scripts.utils import remove_numbers, remove_stop_words, lemmatize


def difflib_overlap(word_token1: list, word_token2: list) -> float:
    """Get similarity percentage from matching sequences between two strings"""

    seq = difflib.SequenceMatcher(a=word_token1, b=word_token2)

    # Return similarity percentage based on difflib library Sequence Matcher
    return round(seq.ratio() * 100, 3)


def calculate_overlap(word_token1: list, word_token2: list) -> float:
    """Get similarity percentage from usage of similar words in two strings"""

    overlapping_words = []

    for word in word_token1:
        if word in word_token2:
            overlapping_words.append(word)

    overlap_percentage = len(overlapping_words) / len(word_token1) * 100

    return round(overlap_percentage, 3)


def calculate_jaccard(word_tokens1: list, word_tokens2: list) -> float:
    """Calculates intersection over union and return Jaccard similarity score"""

    list1, list2 = remove_numbers(word_tokens1), remove_numbers(word_tokens2)
    list1, list2 = remove_stop_words(list1), remove_stop_words(list2)
    list1, list2 = lemmatize(list1), lemmatize(list2)

    # Combine both tokens to find union
    both_tokens = list1 + list2
    union = set(both_tokens)

    # Calculate intersection
    intersection = set()
    for word in list1:
        if word in list2:
            intersection.add(word)

    jaccard_score = len(intersection) / len(union)

    return round(jaccard_score, 3)



""" This script stores useful functions for html_writing.py file

It finds occurences of matching blocks in text.
It gets ordered positions of matching blocks in text.
It returns colors depending on the similarity score.

"""

import difflib
from operator import itemgetter
from typing import List, Tuple
from os import getcwd, path, makedirs


def get_real_matching_blocks(words_list1: list, words_list2: list, minimum_size: int = 2) -> list:
    """Return list of matching blocks with size greater than n"""

    matching_blocks = difflib.SequenceMatcher(a=words_list1, b=words_list2).get_matching_blocks()
    if minimum_size and minimum_size > 0:
        return [b for b in matching_blocks if b.size >= minimum_size]

    return [b for b in matching_blocks if b.size >= 2]


def get_ordered_blocks_positions(string: str, matching_blocks: list, string_blocks: list) -> list:
    """Return ordered list of all positions of matching blocks in string"""

    all_blocks_positions: List[Tuple[int, int]] = []

    for block_ind, _ in enumerate(matching_blocks):
        # Find all positions of substring in string
        block_positions = [char for char in range(len(string)) if string.startswith(string_blocks[block_ind], char)]

        for position in block_positions:
            # We check if there is another block starting at the same position
            var = [pos_tuple for pos_tuple in all_blocks_positions if pos_tuple[0] == position]
            if var:  # If there is one such block
                size = len(string_blocks[var[0][1]])  # get size of block in var
                if size < len(string_blocks[block_ind]):
                    # Remove old position block which is smaller than curr block
                    all_blocks_positions.pop(all_blocks_positions.index(var[0]))

                    # Add new block to list
                    all_blocks_positions.append((position, block_ind))
            else:
                all_blocks_positions.append((position, block_ind))

    return sorted(all_blocks_positions, key=itemgetter(0))


def blocks_list_to_strings_list(blocks_list: list, curr_text: list) -> list:
    """Convert blocks list to len of blocks strings"""

    strings_len_list = []

    for block in blocks_list:
        # Append size of block in string
        strings_len_list.append(len(" ".join(map(str, curr_text[block.a : block.a + block.size]))))

    return strings_len_list


def writing_results(dir_name: str) -> str:
    """Create new directory for results in current working directory"""

    curr_directory = path.dirname(getcwd())
    final_directory = path.join(curr_directory, r"results/" + dir_name)
    if not path.exists(final_directory):
        makedirs(final_directory)

    return final_directory


def get_color_from_similarity(similarity_score: float) -> str:
    """Return css style according to similarity score"""

    if float(similarity_score) > 15:
        return "#990033; font-weight: bold"
    if float(similarity_score) > 10:
        return "#ff6600"
    if float(similarity_score) > 5:
        return "#ffcc00"

    return "green"



import unittest
from unittest.mock import patch
import nltk
from scripts.utils import (
    parse_options,
    is_float,
    wait_for_file,
    remove_numbers,
    remove_stop_words,
    lemmatize,
)


class TestUtils(unittest.TestCase):
    """
    Tests utils.py
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the test class.
        """
        # Download NLTK stopwords
        nltk.download("stopwords")

    def test_parse_options(self):
        """
        Tests parse_options()
        """
        # Mock the arguments and test the parse_options function
        test_args = ["program", "input_dir", "-o", "output_dir", "-s", "5"]
        with patch("sys.argv", test_args):
            args = parse_options()
            self.assertEqual(args.in_dir, "input_dir")
            self.assertEqual(args.out_dir, "output_dir")
            self.assertEqual(args.block_size, 5)

    def test_is_float(self):
        """
        Tests is_float()
        """
        # Test cases for is_float function
        self.assertTrue(is_float(3.14))
        self.assertFalse(is_float(-1))
        self.assertFalse(is_float("not a float"))

    def test_wait_for_file(self):
        """
        Tests wait_for_file()
        """
        # Test the wait_for_file function with a mock path.isfile
        with patch("os.path.isfile", return_value=True):
            self.assertTrue(wait_for_file("dummy_file"))

    def test_remove_numbers(self):
        """
        Tests remove_numbers()
        """
        # Test the remove_numbers function
        words_list = ["hello", "world", 123, 4.56]
        expected = ["hello", "world"]
        self.assertEqual(remove_numbers(words_list), expected)

    def test_remove_stop_words(self):
        """
        Tests remove_stop_words()
        """
        # Test the remove_stop_words function
        with patch("nltk.corpus.stopwords.words", return_value=["a", "the"]):
            words_list = ["a", "quick", "brown", "fox"]
            expected = ["quick", "brown", "fox"]
            self.assertEqual(remove_stop_words(words_list), expected)

    def test_remove_stop_words_only_stopwords(self):
        """
        Tests remove_stop_words()
        """
        # Test with only stopwords
        with patch("nltk.corpus.stopwords.words", return_value=["a", "the"]):
            words_list = ["a", "the"]
            expected = []
            self.assertEqual(remove_stop_words(words_list), expected)

    def test_remove_stop_words_no_stopwords(self):
        """
        Tests remove_stop_words()
        """
        # Test with no stopwords
        with patch("nltk.corpus.stopwords.words", return_value=["a", "the"]):
            words_list = ["quick", "brown", "fox"]
            expected = ["quick", "brown", "fox"]
            self.assertEqual(remove_stop_words(words_list), expected)

    def test_lemmatize(self):
        """
        Tests lemmatize()
        """
        # Test the lemmatize function
        with patch("nltk.stem.WordNetLemmatizer.lemmatize", side_effect=lambda x: x + "_lemmatized"):
            words_list = ["running", "jumps"]
            expected = ["running_lemmatized", "jumps_lemmatized"]
            self.assertEqual(lemmatize(words_list), expected)



#!/usr/bin/env python
""" This module launches the files comparison process

This modules compares all txt, docs, odt, pdf files present in path specified as argument.
It writes results in a HTML table.
It uses difflib library to find matching sequences.
It can also use Jaccard Similarity, words counting, overlapping words for similarity

"""
import sys
import webbrowser
from datetime import datetime
from os import listdir, path
from typing import List

from scripts.html_writing import add_links_to_html_table, results_to_html, papers_comparison
from scripts.html_utils import writing_results
from scripts.processing_files import file_extension_call
from scripts.similarity import difflib_overlap
from scripts.utils import wait_for_file, get_student_names, parse_options


def main() -> None:
    """
    Main function to process and compare text files.

    Parses command-line arguments to obtain input and output directories and block size for comparison.
    Validates the input directory and checks if there are at least two files for comparison.
    Processes each file in the input directory, extracting text and handling different file formats.
    Calculates similarity scores between each pair of processed files using difflib.
    Generates and writes HTML files with colored comparison results in the specified output directory.
    Creates a summary results HTML file with links to individual comparisons and opens it in a web browser.
    Exits the program if the specified path does not exist, or if there are fewer than two files for comparison.
    """

    args = parse_options()
    in_dir, out_dir, block_size = args.in_dir, args.out_dir, args.block_size

    if path.exists(in_dir):  # Check if specified path exists
        if not path.isabs(in_dir):
            in_dir = path.abspath(in_dir)
        if len(listdir(in_dir)) > 1:  # Check if there are at least 2 files at specified path
            filenames, processed_files = [], []
            students_names = get_student_names(in_dir)
            for ind, direc in enumerate(listdir(in_dir)):
                if path.isdir(path.join(in_dir, direc)):
                    for file in listdir(path.join(in_dir, direc)):
                        file_words = file_extension_call(str(path.join(in_dir, direc, file)))

                        if file_words:  # If all files have supported format
                            processed_files.append(file_words)
                            filenames.append(students_names[ind])
                        else:  # At least one file was not supported
                            print("Remove files which are not txt, pdf, docx or odt and run the script again.")
                            sys.exit()
            if out_dir is not None and path.exists(out_dir):
                if not path.isabs(out_dir):
                    out_dir = path.abspath(out_dir)
                results_directory = out_dir
            else:
                # Create new directory for storing html files
                results_directory = writing_results(datetime.now().strftime("%Y%m%d_%H%M%S"))

            difflib_scores: List[List[float]] = [[] for _ in range(len(processed_files))]
            file_ind = 0

            for i, text in enumerate(processed_files):
                for j, text_bis in enumerate(processed_files):
                    if i != j:
                        # Append to the list the similarity score between text and text_bis
                        difflib_scores[i].append(difflib_overlap(text, text_bis))

                        # Write text with matching blocks colored in results directory
                        papers_comparison(
                            results_directory,
                            file_ind,
                            text,
                            text_bis,
                            (filenames[i], filenames[j]),
                            block_size,
                        )
                        file_ind += 1
                    else:
                        difflib_scores[i].append(-1)

            results_directory = path.join(results_directory, "_results.html")
            print(results_directory)

            results_to_html(difflib_scores, filenames, results_directory)

            if wait_for_file(results_directory, 60):  # Wait for file to be created
                add_links_to_html_table(results_directory)
                webbrowser.open(results_directory)  # Open results HTML table
            else:
                print("Results file was not created...")
        else:
            print("Minimum number of files is not present. Please check that there are at least two files to compare.")
            sys.exit()
    else:
        print("The specified path does not exist : " + in_dir)
        sys.exit()


if __name__ == "__main__":
    main()



import subprocess
import logging
from setuptools import setup, find_packages


# Configure logging
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")


def get_version(default="0.1.0"):
    try:
        version = subprocess.check_output(["git", "describe", "--tags", "--long"]).strip().decode("utf-8")
        logging.info(f"Original version from git: {version}")

        if "-" in version:
            # Example output: v0.1.0-3-gaf8ca44
            # Convert to PEP 440 compliant version without local version identifiers
            parts = version.split("-")
            base_version = parts[0]  # 'v0.1.0'
            commit_count = parts[1]  # '3'

            # Remove leading 'v' and split version components
            major, minor, patch = map(int, base_version.lstrip("v").split("."))

            # Increment patch version by commit count
            patch += int(commit_count)
            version = f"{major}.{minor}.{patch}"
            logging.info(f"Normalized version for PyPI: {version}")
    except Exception:
        logging.error("Failed to get version from git, using default version.", exc_info=True)
        version = default
    return version


setup(
    name="copy-spotter",
    version=get_version(),
    packages=find_packages(),
    install_requires=[
        "beautifulsoup4==4.10.0",
        "nltk==3.6.6",
        "odfpy==1.4.1",
        "pdfplumber==0.5.28",
        "slate3k==0.5.3",
        "tabulate==0.8.9",
    ],
    extras_require={
        "lint": ["pylint==3.0.2", "mypy==1.7.1", "flake8==6.1.0", "black==24.3.0", "types-tabulate"],
        "dev": ["pytest", "pre-commit"],
    },
    author="Clément Delteil",
    author_email="clement45.delteil45@gmail.com",
    description="Make plagiarism detection easier. This package will find similar sentences between given files and "
    "highlight them in a side by side comparison.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Wazzabeee/copy-spotter",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
    ],
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "copy-spotter=scripts.main:main",
        ],
    },
)
